import falcon
from .config import get_config, get_error_codes


# Override `falcon.HTTPError` to include error messages in responses.
class HTTPError(falcon.HTTPError):
    """
    HTTPError that stores a dictionary of validation error messages.
    """

    def __init__(self, status, errors=None, *args, **kwargs):
        self.errors = errors
        super().__init__(status, *args, **kwargs)

    def to_dict(self, *args, **kwargs):
        ret = super().to_dict(*args, **kwargs)
        if self.errors is not None:
            ret['errors'] = self.errors
        return ret


class JsonRpcError(falcon.HTTPError):

    def __init__(self, code: int, description: str = None, status: int = falcon.status.HTTP_200, *args, **kwargs):
        super().__init__(status, code=code, description=description, *args, **kwargs)


class JsonDecodingError(JsonRpcError):

    def __init__(self, description: str = 'JSON decoding failed', *args, **kwargs):
        super().__init__(code=get_error_codes().json_decoding, description=description, *args, **kwargs)


class JsonEncodingError(JsonRpcError):

    def __init__(self, description: str = 'JSON encoding failed', *args, **kwargs):
        super().__init__(code=get_error_codes().json_decoding, description=description, *args, **kwargs)


class JsonRpcSchemaError(JsonRpcError):

    def __init__(self, description: str = 'JSON RPC request validation failed', *args, **kwargs):
        super().__init__(code=get_error_codes().json_rpc_schema_validation, description=description, *args, **kwargs)


class JsonRpcParamsError(JsonRpcError):

    def __init__(self, description: str = 'JSON RPC params validation failed', *args, **kwargs):
        super().__init__(code=get_error_codes().json_rpc_params_validation, description=description, *args, **kwargs)


class JsonRpcMethodNotFoundError(JsonRpcError):

    def __init__(self, description: str = 'This JSON RPC method is not available', *args, **kwargs):
        super().__init__(code=get_error_codes().json_rpc_method_not_found, description=description, *args, **kwargs)


class JsonRpcMethodCallError(JsonRpcError):

    def __init__(self, description: str = 'Internal error. JSON RPC method call failed.', *args, **kwargs):
        super().__init__(code=get_error_codes().json_rpc_method_call, description=description, *args, **kwargs)




from .config import InjectedAttributes


class JsonAPIExchange:
    def __init__(self, rpc: bool = True):
        self.rpc = rpc

    def __call__(self, cls):
        class WrappedJsonExchange(cls):
            pass

        setattr(WrappedJsonExchange, InjectedAttributes.json_exchange_flag, True)
        setattr(WrappedJsonExchange, InjectedAttributes.enable_json_rpc, self.rpc)
        return WrappedJsonExchange


JsonRpc = JsonAPIExchange(rpc=True)
JsonExchange = JsonAPIExchange(rpc=False)



import jsonschema
from typing import Any
from .errors import JsonRpcSchemaError

JsonRpcRequestSchema = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "A JSON RPC 2.0 request",
    "oneOf": [
        {
            "description": "An individual request",
            "$ref": "#/definitions/request"
        },
        {
            "description": "An array of requests",
            "type": "array",
            "items": {"$ref": "#/definitions/request"},
            "minItems": 1
        }
    ],
    "definitions": {
        "request": {
            "type": "object",
            "required": ["jsonrpc", "method"],
            "properties": {
                "jsonrpc": {"enum": ["2.0"]},
                "method": {
                    "type": "string"
                },
                "id": {
                    "type": ["string", "number"]
                },
                "params": {
                    "type": ["array", "object"]
                }
            }
        }
    }
}


def validate_json_rpc_data(data: Any, schema: Any, exception_cls: type = JsonRpcSchemaError):
    try:
        jsonschema.validate(data, schema, format_checker=jsonschema.FormatChecker())
    except jsonschema.ValidationError as e:
        raise exception_cls(e.message)


def validate_json_rpc_request(req):
    if not 'json' in req.context:
        raise JsonRpcSchemaError('Context does not contain `json`. '
                                 'Perhaps, middleware was not initialized')
    validate_json_rpc_data(req.context['json'], JsonRpcRequestSchema, JsonRpcSchemaError)



import inspect
import types
import re
from typing import Dict, Any, Union, List, Callable
from falcon import Request, HTTPError
from functools import partial
from .config import config, InjectedAttributes
from .util import json_encode
from .validators import validate_json_rpc_data
from .errors import JsonRpcParamsError, JsonRpcMethodCallError, JsonRpcError
from logging import getLogger
import pkgutil

logger = getLogger(__name__)

JsonRpcParams = Union[None, Dict[str, Any], List[Any]]


def _register_entrypoint(fn, entrypoint):
    descriptors = getattr(fn, InjectedAttributes.entry_point, None)
    if descriptors is None:
        descriptors = set()
        setattr(fn, InjectedAttributes.entry_point, descriptors)
    descriptors.add(entrypoint)


def json_rpc_entrypoint_container_extractor(module):
    def is_entrypoint(method):
        return hasattr(method, InjectedAttributes.entry_point)

    containers = list()
    for _, maybe_container in inspect.getmembers(module, lambda obj: isinstance(obj, type)):
        if inspect.getmembers(maybe_container, is_entrypoint):
            containers.append(maybe_container)
    return containers


class JsonRpcEntryPoint:
    name = None
    method_name = None
    router = None
    method_container = None
    validation_schema = None
    validation_error_code = None
    __params = None

    def __new__(cls, *args, **kwargs):
        inst = super().__new__(cls)
        inst.__params = (args, kwargs)
        return inst

    def __init__(self, name: str, **kwargs):
        self.name = name
        if 'schema' in kwargs:
            self.validation_schema = kwargs['schema']

    def call(self, params: Union[Dict[str, Any], List[Any]] = None):
        args = []
        kwargs = {}
        if params:
            if type(params) is list:
                args = params
            else:
                kwargs = params

        if self.validation_schema:
            if len(args):
                validate_json_rpc_data(list(args), self.validation_schema, JsonRpcParamsError)
            elif len(kwargs):
                validate_json_rpc_data(kwargs, self.validation_schema, JsonRpcParamsError)
        self.check_signature(args, kwargs)
        try:
            return getattr(self.method_container, self.method_name)(*args, **kwargs)
        except JsonRpcError:
            raise
        except Exception as e:
            err_descr = f'Internal error during calling method `{self.method_name}`.'
            logger.error(err_descr, exc_info=True)
            raise JsonRpcMethodCallError(err_descr)

    def bind(self, router, method_container, method_name):
        def clone(prototype):
            if prototype.is_bound():
                raise RuntimeError(f'Cannot bind already bound entry point `{method_name}`.')

            cls = type(prototype)
            args, kwargs = prototype.__params
            instance = cls(*args, **kwargs)
            instance.router = router
            try:
                instance.method_container = method_container()
            except Exception as e:
                if str(e).__contains__('required positional argument'):
                    raise Exception(f'Class-container of JSON RPC method `{method_name}` '
                                    'must not have required initialization arguments.')
                # instantiation raised an exception, throw it as it is
                raise

            return instance

        instance = clone(self)
        instance.method_name = method_name
        return instance

    def is_bound(self):
        return self.method_container is not None

    def check_signature(self, args, kwargs):
        fn = getattr(self.method_container, self.method_name)
        try:
            inspect.getcallargs(fn, *args, **kwargs)
        except TypeError as exc:
            raise JsonRpcParamsError(f'Wrong method signature. {str(exc)}')
        except Exception as exc:
            raise JsonRpcParamsError(f'Checking method signature failed. {str(exc)}')

    @classmethod
    def decorator(cls, *decoargs, **decokwargs):
        """
        Decorator for JSON RPC methods. Only instance methods can be decorated.

        :param decoargs: If first argument is provided, it is used as method name.
        :param decokwargs: If 'name' key is provided, it is used as method name.
        :return:
        """

        def wrapped_f(fn, args, kwargs):
            name = decokwargs.get('name', fn.__name__)
            if not (len(decoargs) == 1 and isinstance(decoargs[0], types.FunctionType)):
                if len(args) >= 1:
                    name = args[0]
                    args = args[1:]

            instance = cls(name, *args, **kwargs)
            descriptors = getattr(fn, InjectedAttributes.entry_point, None)
            if not descriptors:
                descriptors = set()
                setattr(fn, InjectedAttributes.entry_point, descriptors)
            descriptors.add(instance)
            return fn

        if len(decoargs) == 1 and isinstance(decoargs[0], types.FunctionType):
            return wrapped_f(decoargs[0], args=(), kwargs={})
        else:
            return partial(wrapped_f, args=decoargs, kwargs=decokwargs)


class JsonRpcResponseError:

    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message


class JsonRpcRequest:

    def __init__(self, req: Request = None):
        a = InjectedAttributes.context_json_data
        json = req.context[a] if req and hasattr(req, 'context') and a in req.context else {}
        self.jsonrpc: str = json.get('jsonrpc', None)
        self.id: str = json.get('id', None)
        self.method: str = json.get('method', None)
        self.params: JsonRpcParams = json.get('params', None)
        self.original_request: Request = req


class JsonRpcResponse:

    def __init__(self, req: Request, result: Any = None, error: JsonRpcResponseError = None):
        self._request_data: JsonRpcRequest = self._get_request_data(req)
        self.jsonrpc: str = self._request_data.jsonrpc if self._request_data else config.json_rpc_version
        self.id: str = self._request_data.id if self._request_data else None
        self.result: Any = result
        self.error: JsonRpcResponseError = error

    def to_json(self, encode=True):
        o = {'jsonrpc': self.jsonrpc, 'id': self.id}
        if self.error is not None:
            o['error'] = {'code': self.error.code, 'message': self.error.message}
        if self.result is not None and self.error is None:
            o['result'] = self.result
        if self.result is None and self.error is None:
            o['result'] = None
        if not encode:
            return o
        return json_encode(o)

    def _get_request_data(self, req: Request) -> JsonRpcRequest:
        a = InjectedAttributes.context_json_data
        return req.context[a] if req and hasattr(req, 'context') and a in req.context else None


JsonRpcMethod = JsonRpcEntryPoint.decorator



from falcon import Response, Request, HTTP_200
from typing import Any
from .json_rpc import JsonRpcRequest, JsonRpcResponse
from .errors import JsonDecodingError
from .validators import validate_json_rpc_request
from .util import json_encode, update_config
from .config import config, InjectedAttributes
import json
# from .logging import logger
from logging import getLogger

logger = getLogger(__name__)


def _read_json_stream(req: Request) -> Any:
    req.context[InjectedAttributes.context_json_data] = {}
    if req.content_length:
        raw = None
        try:
            raw = req.stream.read().strip()
            req.context[InjectedAttributes.context_json_data] = json.loads(raw.decode('utf-8'))
        except UnicodeDecodeError:
            raise JsonDecodingError('Wrong request encoding. Must be UTF-8.')
        except ValueError:
            logger.exception(f'JSON syntax error, raw data: {raw}.')
            raise JsonDecodingError(f'JSON syntax error.')
        except BaseException as e:
            logger.exception(f'JSON parse unknown error.')
            raise JsonDecodingError(f'JSON parse error.')


def _is_rpc_enabled(resource) -> bool:
    return getattr(resource, InjectedAttributes.enable_json_rpc, False)


def _is_json_exchange(resource) -> bool:
    return getattr(resource, InjectedAttributes.json_exchange_flag, False)


class JsonRpcMiddleware(object):

    def __init__(self, **kwargs):
        update_config(**kwargs)

    def process_resource(self, req, resp, resource, params):
        if _is_json_exchange(resource):
            _read_json_stream(req)
            if _is_rpc_enabled(resource):
                if req.method.lower() == 'post':
                    if config.auto_validate_json_rpc_request:
                        validate_json_rpc_request(req)
                req.context[InjectedAttributes.context_json_data] = JsonRpcRequest(req)

    def process_response(self, req: Request, resp: Response, resource: Any, req_succeeded: bool):
        # logger.debug(f'Response is {resp.__dict__}')
        # logger.debug(f'Response body is {resp.body}')
        json_result = getattr(resp, InjectedAttributes.context_json_data, None)
        if json_result is not None:
            if _is_rpc_enabled(resource):
                resp.body = JsonRpcResponse(req, result=json_result).to_json(False)
            else:
                resp.body = json_result
            # logger.debug(resp.body)
            resp.body = json_encode(resp.body)


class CorsMiddleware:

    def __init__(self, allow_origin: str = '*', max_age: int = 86400):
        self.allow_origin = allow_origin
        self.max_age = max_age

    def process_response(self, req, resp, resource, req_succeeded):
        if not self.allow_origin or not len(self.allow_origin):
            return
        resp.set_header('Access-Control-Allow-Origin', self.allow_origin)
        if (req_succeeded
                and req.method == 'OPTIONS'
                and req.get_header('Access-Control-Request-Method')
        ):
            # NOTE: This is a CORS preflight request. Patch the response accordingly.
            allow = resp.get_header('Allow')
            resp.delete_header('Allow')
            allow_headers = req.get_header('Access-Control-Request-Headers', default='*')
            resp.set_headers((
                ('Access-Control-Allow-Methods', allow),
                ('Access-Control-Allow-Headers', allow_headers),
                ('Access-Control-Max-Age', self.max_age),
            ))



import falcon
from typing import Dict, Any, List
from .json_rpc_router import JsonRpcRouter
from .middleware import CorsMiddleware, JsonRpcMiddleware
from .util import update_config
from .config import get_config
from logging import getLogger

logger = getLogger(__name__)


class JsonRpcServer:

    def __init__(self, routes: Dict[str, str] = None,
                 config: Dict[str, Any] = None,
                 middleware: List[Any] = None) -> falcon.API:
        """

        :param routes: Dictionary of JSON RPC API routes in format {`route`: `methods_module_path`}
        :param config: Global configuration, check available options in `config.py`
        :param middleware: List of middleware to be passed to `falcon` application
        """
        self.falcon: falcon.API = None
        self.middleware: List[Any] = middleware if middleware else list()
        self.routers: Dict[str, JsonRpcRouter] = {}
        self.routes: Dict[str, str] = routes or {}

        if config:
            update_config(**config)
        self.config = get_config()
        self._setup_falcon()

    def _setup_falcon(self):
        add_json_rpc_middleware = True
        add_cors_middleware = self.config.access_control_allow_origin and len(self.config.access_control_allow_origin)

        for m in self.middleware:
            if type(m) is JsonRpcMiddleware:
                add_json_rpc_middleware = False
                logger.warning(f'{type(m).__name__} was initialized in main application. '
                               'This step is not required, it would be passed to falcon anyway.')
            if type(m) is CorsMiddleware:
                add_cors_middleware = False
                logger.warning(f'{type(m).__name__} was initialized in main application. '
                               'This step is not required, instead you could use configuration '
                               'options `access_control_allow_origin` amd/or `access_control_max_age`')

        if add_json_rpc_middleware:
            self.middleware.append(JsonRpcMiddleware())
        if add_cors_middleware:
            self.middleware.append(CorsMiddleware(self.config.access_control_allow_origin,
                                                  self.config.access_control_max_age))
        self.falcon = falcon.API(middleware=self.middleware)
        self.bind_routes(self.routes)

    def bind_routes(self, routes: Dict[str, str]):
        for route, methods_path in routes.items():
            logger.debug(f'Binding route `{route}` to `{methods_path}`.')
            self.routers[route] = JsonRpcRouter(self.falcon, route, methods_path)

    @property
    def application(self):
        return self.falcon



from .server import JsonRpcServer
from .middleware import JsonRpcMiddleware, CorsMiddleware
from .json_rpc import JsonRpcMethod
from .errors import JsonRpcError
from .json_rpc_router import JsonRpcRouter
from .decorators import JsonExchange, JsonRpc

__author__ = 'Sergii Bibikov'
__email__ = 'sergeport@gmail.com'

name = 'falcon_json_rpc'



from datetime import datetime
from typing import Any
import json
from .config import config
from .errors import JsonEncodingError


class JsonDateTimeEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return json.JSONEncoder.default(self, o)


def json_encode(data: Any):
    try:
        return str.encode(json.dumps(data, cls=JsonDateTimeEncoder))
    except Exception as e:
        raise JsonEncodingError(f'JSON encoding failed. {e}')


def update_config(**kwargs):
    for k in kwargs:
        if hasattr(config, k):
            setattr(config, k, kwargs.get(k))








from falcon import Response, Request, API
from typing import Callable, Dict, Any, List
from .errors import JsonRpcError, JsonRpcMethodNotFoundError, JsonRpcMethodCallError
from muhomor.utils.filesystem import decorator_containers_from_module
from .json_rpc import (
    json_rpc_entrypoint_container_extractor,
    JsonRpcRequest,
    JsonRpcParams
)
import inspect
import os
from logging import getLogger
from .serializers import error_serializer
from .config import InjectedAttributes
from .decorators import JsonRpc

logger = getLogger(__name__)


@JsonRpc
class JsonRpcRouter:

    def __init__(self, application: API, entry_point: str, methods_path: str):
        self._setup(application, entry_point, methods_path)
        self._setup_application()

    def _setup(self, application: API, entry_point: str, methods_path: str):
        self.application = application
        self.methods: Dict[str, Callable] = {}
        self.registered_method_names: List[str] = list()
        self.entry_point = entry_point
        self.methods_path = methods_path
        self.method: str = None
        self.params: JsonRpcParams = None
        self.data: JsonRpcRequest = JsonRpcRequest()
        self._register_methods(self.methods_path)

    def _setup_application(self):
        self.application.set_error_serializer(error_serializer)
        self.application.add_route(self.entry_point, self)

    def _register_methods(self, methods_path: str):
        method_containers = decorator_containers_from_module(methods_path, json_rpc_entrypoint_container_extractor)
        for container in method_containers:
            for method_name, method in inspect.getmembers(container, inspect.isfunction):
                entrypoints = getattr(method, InjectedAttributes.entry_point, [])
                for entrypoint in entrypoints:
                    if entrypoint.name in self.methods:
                        raise Exception(f'JSON-RPC method {entrypoint.name} is registered already')
                    bound = entrypoint.bind(self, container, method_name)
                    setattr(container, 'resource', self)
                    self.methods[entrypoint.name.lower()] = bound.call
                    self.registered_method_names.append(entrypoint.name.lower())

    def _call(self, method, params):
        logger.debug(f'JSON RPC router methods in PID {os.getpid()}: {self.registered_method_names}')
        if method not in self.registered_method_names:
            raise JsonRpcMethodNotFoundError(f'Method {method} is not available')
        return self.methods[method](params)

    def on_post(self, req: Request, resp: Response):
        try:
            j: JsonRpcRequest = req.context[InjectedAttributes.context_json_data]
            result = self._call(j.method.lower(), j.params)
            setattr(resp, InjectedAttributes.context_json_data, result)
        except JsonRpcError:
            logger.exception('Error `JsonRpcError` calling JSON RPC method')
            raise
        except BaseException:
            logger.exception('Error calling JSON RPC method')
            raise

    def on_get(self, req: Request, resp: Response):
        pass

    def on_put(self, req: Request, resp: Response):
        pass

    def on_patch(self, req: Request, resp: Response):
        pass

    def on_options(self, req: Request, resp: Response):
        pass



from logging import getLogger
from .json_rpc import JsonRpcResponse, JsonRpcResponseError
from .errors import JsonRpcError

logger = getLogger(__name__)

def error_serializer(req, resp, exception):
    if isinstance(exception, JsonRpcError):
        logger.warning(f'Serializing `JsonRpcError` {exception} {exception.__dict__} into resp.body')
        e = JsonRpcResponseError(exception.code, exception.description)
        inst = JsonRpcResponse(req=req, error=e)
        resp.body = inst.to_json()
    else:
        logger.warning(f'Serializing `UnknownError` {exception} {exception.__dict__} into resp.http_error')
        resp.http_error = exception
    resp.content_type = 'application/json'
    # resp.append_header('Access-Control-Allow-Origin', '*')



# TO-DO: ACCESS CONFIG ONNLY THROUGH GET_CONFIG


class Config:

    def __init__(self):
        self.json_rpc_version: str = "2.0"
        self.enable_json_rpc: bool = True
        self.auto_validate_json_rpc_request: bool = True

        # set to None or empty string do disable
        self.access_control_allow_origin: str = None
        # in seconds
        self.access_control_max_age: int = 86400  # 24 hours


class ErrorCodes:

    def __init__(self):
        self.json_decoding: int = -77001
        self.json_encoding: int = -77002
        self.json_rpc_schema_validation: int = -77003
        self.json_rpc_params_validation: int = -77004
        self.json_rpc_unavailable_method: int = -77005
        self.json_rpc_method_call: int = -77006
        self.json_rpc_method_not_found: int = -77007


class InjectedAttributes:
    entry_point: str = 'json_rpc_methods'
    context_json_data: str = 'json'
    json_exchange_flag: str = 'json_exchange'
    enable_json_rpc: str = 'enable_json_rpc'


config = Config()

error_codes = ErrorCodes()


def get_error_codes() -> ErrorCodes:
    return error_codes


def get_config() -> Config:
    return config
